@@ -1,15 +1,44 @@
 from pikachu.reactions.functional_groups import GroupDefiner, find_atoms, find_bonds
-from pikachu.general import read_smiles
-from raichu.data.attributes import ATTRIBUTES
 from raichu.reactions.general import initialise_atom_attributes
-from raichu.attach_to_domain import attach_to_domain_pk
+
+import timeout_decorator
 from raichu.data.molecular_moieties import AMINO_FATTY_ACID, AMINO_ACID_BACKBONE, N_AMINO_ACID, C1_AMINO_ACID, \
     C2_AMINO_ACID, BETA_AMINO_ACID_BACKBONE, B_N_AMINO_ACID, B_C1_AMINO_ACID, B_C2_AMINO_ACID, B_C3_AMINO_ACID, \
-    B_LEAVING_BOND, ACID_C1
+    B_LEAVING_BOND, ACID_C1, OH_AMINO_ACID, N_AMINO_ACID_ATTACHED, C1_AMINO_ACID_ATTACHED, C2_AMINO_ACID_ATTACHED
 
 POLYKETIDE_S = GroupDefiner('Sulphur atom polyketide', 'SC(C)=O', 0)
+POLYKETIDE_S_INSERTED_O = GroupDefiner('Sulphur atom polyketide inserted O', 'SC(O)=O', 0)
+POLYKETIDE_S_REDUCED_STARTER = GroupDefiner('Sulphur atom polyketide', 'SC(C)O', 0)
 
 
+class CentralChain:
+    def __init__(self, structure, chain_type):
+        self.structure = structure
+        if chain_type in ['ripp', 'nrps', 'pks']:
+            self.chain_type = chain_type
+        else:
+            raise ValueError(f"Expected 'ripp', 'nrps', or 'pks' as central chain type; got {chain_type}")
+
+    def label_central_chain(self):
+        raise NotImplementedError
+
+
+class PksCentralChain(CentralChain):
+    def __init__(self, structure):
+        super().__init__(structure, chain_type='pks')
+
+
+class NrpsCentralChain(CentralChain):
+    def __init__(self, structure):
+        super().__init__(structure, chain_type='nrps')
+
+
+class RippCentralChain(CentralChain):
+    def __init__(self, structure):
+        super().__init__(structure, chain_type='ripp')
+
+
+@timeout_decorator.timeout(3)
 def label_pk_central_chain(pk_starter_unit):
     """Finds the the atoms in the central chain of the polyketide starter unit,
     sets the in_central_chain Atom object attribute to True/False accordingly,
@@ -24,7 +53,12 @@ def label_pk_central_chain(pk_starter_unit):
 
     # Check if structure is a polyketide, define sulphur attached to domain
     locations_sulphur = find_atoms(POLYKETIDE_S, pk_starter_unit)
+    if len(locations_sulphur) == 0:
+        locations_sulphur = find_atoms(POLYKETIDE_S_INSERTED_O, pk_starter_unit)
+        if len(locations_sulphur) == 0:
+            locations_sulphur = find_atoms(POLYKETIDE_S_REDUCED_STARTER, pk_starter_unit)
     assert len(locations_sulphur) == 1
+    
     sulphur_polyketide = locations_sulphur[0]
     central_chain.append(sulphur_polyketide)
 
@@ -38,13 +72,16 @@ def label_pk_central_chain(pk_starter_unit):
             visited.append(chain_carbon)
 
             while not end_carbon:
+
                 for next_atom in chain_carbon.neighbours:
                     ethyl_branch = False
                     inside_cycle = False
                     methyl_group = False
 
                     # Build lists of neighbouring atom types
-                    if next_atom.type == 'C' and next_atom not in visited:
+                    if next_atom.type == 'C' and next_atom not in visited or \
+                            (next_atom.type == 'O' and next_atom not in visited and
+                             len(next_atom.get_neighbours('C')) == 2):
 
                         # Keep track of carbon neighbours
 
@@ -53,15 +90,15 @@ def label_pk_central_chain(pk_starter_unit):
                         c_neighbours = next_atom.get_neighbours('C')
                         if len(c_neighbours) == 2 and next_atom.get_neighbour('O') is None:
 
-                            for atom in c_neighbours:
-                                for neighbour in atom.neighbours:
-                                    types.append(neighbour.type)
+                            for carbon_neighbour in c_neighbours:
+                                for branch_neighbour in carbon_neighbour.neighbours:
+                                    types.append(branch_neighbour.type)
 
                             if types.count('H') == 4 and types.count('C') == 4:
                                 ethyl_branch = True
 
-                            for atom in c_neighbours:
-                                if atom not in visited and atom != next_atom:
+                            for carbon_neighbour in c_neighbours:
+                                if carbon_neighbour not in visited and carbon_neighbour != next_atom:
                                     ethyl_branch = False
 
                             visited.append(chain_carbon)
@@ -69,14 +106,17 @@ def label_pk_central_chain(pk_starter_unit):
                         # Carbon in a terminal carboxylic acid group is the
                         # final carbon in the central chain
                         if len(next_atom.get_neighbours('O')) == 2:
+                            for oxygen in next_atom.get_neighbours('O'):
+                                if next_atom.get_bond(oxygen).type == 'single':
+                                    central_chain.append(oxygen)
+                                    visited.append(oxygen)
                             central_chain.append(next_atom)
                             end_carbon = True
                             visited.append(chain_carbon)
 
                         # Confirm carbon is not part of cycle
                         if len(c_neighbours) == 2:
-                            if all(atom.in_ring(pk_starter_unit) for atom in c_neighbours):
-                                print(15)
+                            if all(carbon.in_ring(pk_starter_unit) for carbon in c_neighbours):
                                 visited.append(next_atom)
                                 inside_cycle = True
                                 end_carbon = True
@@ -121,13 +161,14 @@ def label_pk_central_chain(pk_starter_unit):
                             for neighbouring_atom in chain_carbon.neighbours:
                                 if neighbouring_atom.type == 'C' and neighbouring_atom not in visited:
                                     count_c_neighbours_not_visited += 1
+
                             if count_c_neighbours_not_visited < 2:
                                 methyl_group = False
                                 visited.append(chain_carbon)
-                                for neighbour in chain_carbon.neighbours:
-                                    if neighbour.type == 'C' and neighbour not in visited:
+                                for carbon_neighbour in chain_carbon.neighbours:
+                                    if carbon_neighbour.type == 'C' and carbon_neighbour not in visited:
                                         next_neighbours = []
-                                        for next_neighbour in neighbour.neighbours:
+                                        for next_neighbour in carbon_neighbour.neighbours:
                                             next_neighbours.append(
                                                 next_neighbour.type)
                                         if next_neighbours.count('H') == 3 or (
@@ -135,7 +176,7 @@ def label_pk_central_chain(pk_starter_unit):
                                                         'H') == 2 and (
                                                 next_neighbours.count(
                                                         '*')) == 1):
-                                            central_chain.append(neighbour)
+                                            central_chain.append(carbon_neighbour)
                                 end_carbon = True
 
                             visited.append(chain_carbon)
@@ -156,6 +197,26 @@ def label_pk_central_chain(pk_starter_unit):
             atom.annotations.in_central_chain = False
 
 
+def label_ripp_central_chain(peptide):
+    initialise_atom_attributes(peptide)
+
+    n_atoms_aa = find_atoms(N_AMINO_ACID_ATTACHED, peptide)
+    c1_atoms_aa = find_atoms(C1_AMINO_ACID_ATTACHED, peptide)
+    c2_atoms_aa = find_atoms(C2_AMINO_ACID_ATTACHED, peptide)
+    oh_atoms_aa = find_atoms(OH_AMINO_ACID, peptide)
+
+    for nitrogen in n_atoms_aa:
+        nitrogen.annotations.in_central_chain = True
+        nitrogen.annotations.n_atom_nmeth = True
+    for carbon in c1_atoms_aa:
+        carbon.annotations.in_central_chain = True
+        carbon.annotations.chiral_c_ep = True
+    for carbon in c2_atoms_aa:
+        carbon.annotations.in_central_chain = True
+    for oxygen in oh_atoms_aa:
+        oxygen.annotations.in_central_chain = True
+
+
 def label_nrp_central_chain(peptide, module_type='elongation'):
     initialise_atom_attributes(peptide)
 
@@ -168,7 +229,6 @@ def label_nrp_central_chain(peptide, module_type='elongation'):
             as_normal = True
 
     if as_normal or module_type == 'elongation':
-
         if peptide.find_substructures(AMINO_ACID_BACKBONE):
             n_atoms_aa = find_atoms(N_AMINO_ACID, peptide)
             c1_atoms_aa = find_atoms(C1_AMINO_ACID, peptide)
@@ -183,6 +243,7 @@ def label_nrp_central_chain(peptide, module_type='elongation'):
             c1_atoms_aa[0].annotations.in_central_chain = True
             c1_atoms_aa[0].annotations.chiral_c_ep = True
             c2_atoms_aa[0].annotations.in_central_chain = True
+
         elif peptide.find_substructures(BETA_AMINO_ACID_BACKBONE):
             n_atoms_aa = find_atoms(B_N_AMINO_ACID, peptide)
             c1_atoms_aa = find_atoms(B_C1_AMINO_ACID, peptide)
@@ -230,7 +291,7 @@ def label_acid_central_chain(acid):
     c2_atom = None
     oh_o_atom = None
     oh_h_atom = None
-    visited = []
+    backbone = []
 
     c1_atoms_aa = find_atoms(ACID_C1, acid)
     if c1_atoms_aa:
@@ -273,102 +334,101 @@ def label_acid_central_chain(acid):
     c2_atom.annotations.in_central_chain = True
     c2_atom.annotations.c2_acid = True
 
-    visited.append(c1_atom)
-    visited.append(c2_atom)
+    backbone.append(c1_atom)
+    backbone.append(c2_atom)
 
-    starting_point = c1_atom
+    last_backbone_atom = c1_atom
     endpoint = False
-    ring_members = 0
-    if starting_point.in_ring(acid):
-        ring_members += 1
 
     # Add atoms in chain with exactly 2 non-H neighbours to the central chain
     while not endpoint:
-        for neighbour in starting_point.get_non_hydrogen_neighbours():
-            neighbour_in_ring = neighbour.in_ring(acid)
-            if neighbour_in_ring:
-                ring_members += 1
-            neighbour_found = False
-            if len(neighbour.get_non_hydrogen_bonds()) == 2 and neighbour.type == 'C' and neighbour not in visited \
-                    and not (ring_members > 2 and neighbour_in_ring):
-                neighbour.annotations.in_central_chain = True
-                visited.append(neighbour)
-                starting_point = neighbour
-                neighbour_found = True
 
-            if not neighbour_found and neighbour.type == 'C' and neighbour not in visited and \
-                    len(neighbour.get_non_hydrogen_bonds()) >= 2 and not (ring_members > 2 and neighbour_in_ring):
-                terminal_count = 0
-                for atom in neighbour.neighbours:
-                    if atom not in visited and len(atom.get_non_hydrogen_bonds()) == 1:
-                        terminal_count += 1
+        neighbours = last_backbone_atom.get_non_hydrogen_neighbours()[:]
+        for neighbour in last_backbone_atom.get_non_hydrogen_neighbours():
+            if neighbour in backbone:
+                neighbours.remove(neighbour)
 
-                if len(neighbour.get_non_hydrogen_bonds()) - terminal_count == 2:
-                    neighbour.annotations.in_central_chain = True
-                    visited.append(neighbour)
-                    starting_point = neighbour
-                    neighbour_found = True
+        if not neighbours:
+            endpoint = True
 
-                if neighbour_found:
-                    break
+        else:
+            # If the last backbone atom was the first ring member
+            if last_backbone_atom.in_ring(acid):
+                ring_members = last_backbone_atom.get_ring(acid)
+                outgoing_pairs = []
+                for atom in ring_members:
+                    for neighbour in atom.get_non_hydrogen_neighbours():
+                        if neighbour not in ring_members and neighbour not in backbone and not neighbour.in_ring(acid):
+                            outgoing_pairs.append((atom, neighbour))
 
-            if not neighbour_found and len(neighbour.get_non_hydrogen_bonds()) == 2 and neighbour not in visited \
-                    and not (ring_members > 2 and neighbour_in_ring):
-                neighbour.annotations.in_central_chain = True
-                visited.append(neighbour)
-                starting_point = neighbour
-                break
+                if not outgoing_pairs:
+                    endpoint = True
 
-            neighbours = []
-            for nb in starting_point.get_non_hydrogen_neighbours():
-                if nb not in visited:
-                    neighbours.append(nb)
+                else:
+                    longest_index = -1
+                    largest_subtree = 0
+                    best_shortest_path = None
+                    for i, pair in enumerate(outgoing_pairs):
+                        ring_atom, outgoing_atom = pair
+                        shortest_path = find_shortest_path_ring(ring_atom, last_backbone_atom, ring_members)
+                        subtree_size = acid.get_subtree_size(outgoing_atom, set(ring_members))
+                        if len(shortest_path) < 4 and subtree_size > largest_subtree:
+                            longest_index = i
+                            best_shortest_path = shortest_path
 
-            if any(atom.in_ring(acid) for atom in neighbours) and neighbour_in_ring:
-                endpoint = True
-                for atom in neighbours:
-                    if len(atom.get_non_hydrogen_bonds()) == 1 and atom.get_non_hydrogen_bonds()[0].type == 'single':
-                        atom.annotations.in_central_chain = True
-                break
+                    if longest_index >= 0:
 
-            elif not any(len(atom.get_non_hydrogen_bonds()) == 2 for atom in neighbours):
+                        backbone_ring_atom, next_backbone_atom = outgoing_pairs[longest_index]
+                        for atom in best_shortest_path:
+                            shortest_path_atom = acid.get_atom(atom)
+                            if shortest_path_atom not in backbone:
+                                shortest_path_atom.annotations.in_central_chain = True
+                                backbone.append(shortest_path_atom)
+                            next_backbone_atom = acid.get_atom(next_backbone_atom)
+                            next_backbone_atom.annotations.in_central_chain = True
+                            backbone.append(next_backbone_atom)
+                            last_backbone_atom = next_backbone_atom
+                    else:
+                        endpoint = True
 
-                endpoint = True
+            else:
+                longest_index = -1
+                largest_subtree = 0
 
-                for atom in neighbours:
-                    terminal_count = 0
-                    for atom_2 in atom.neighbours:
-                        if atom_2 not in visited and len(atom_2.get_non_hydrogen_bonds()) == 1:
-                            terminal_count += 1
+                for i, neighbour in enumerate(neighbours):
+                    subtree_size = acid.get_subtree_size(neighbour, set(backbone))
+                    if subtree_size > largest_subtree:
+                        largest_subtree = subtree_size
+                        longest_index = i
 
-                    if len(atom.get_non_hydrogen_bonds()) - terminal_count == 2:
-                        endpoint = False
-                        break
+                if longest_index >= 0:
+                    next_backbone_atom = acid.get_atom(neighbours[longest_index])
+                else:
+                    next_backbone_atom = acid.get_atom(neighbours[0])
 
-                if endpoint:
-
-                    # Add the last atom in the chain to the central chain
-                    for atom in neighbours:
-                        if len(atom.get_non_hydrogen_bonds()) == 1 and atom.get_non_hydrogen_bonds()[0].type == 'single':
-                            atom.annotations.in_central_chain = True
-                    break
+                next_backbone_atom.annotations.in_central_chain = True
+                backbone.append(next_backbone_atom)
+                last_backbone_atom = next_backbone_atom
 
 
-if __name__ == "__main__":
-    starter_units_antismash = ['SC(=O)CC', 'SC(CC(O)=O)=O', 'SC(CC(O)=O)=O',
-                               'SC(C(C(O)=O)CC)=O', 'SC(C(C(O)=O)OC)=O',
-                               'SC(C*)=O',
-                               'SC(C(C)CC)=O', 'SC(C1C(CCC1)C(=O)O)=O',
-                               'SC(C)=O',
-                               'SC(C1=CC=CC=C1)=O', 'SC(CC(C)C)=O',
-                               'SC(C(C(=O)O)CC[Cl])=O']
-    starter_unit_error = 'SC(C1=CC=CC=C1)=O'
-    struct = read_smiles(starter_unit_error)
-    struct.add_attributes(ATTRIBUTES, boolean=True)
-    struct = attach_to_domain_pk(struct)
-    label_pk_central_chain(struct)
-    struct.print_graph()
-    for atom in struct.graph:
-        print(atom, atom.annotations.in_central_chain)
+def find_shortest_path_ring(atom_1, atom_2, atom_list):
+    max_distance = len(atom_list) // 2
+    atom_1_idx = None
+    atom_2_idx = None
+    for i, atom in enumerate(atom_list):
+        if atom == atom_1:
+            atom_1_idx = i
+        if atom == atom_2:
+            atom_2_idx = i
 
+    assert atom_1_idx is not None
+    assert atom_2_idx is not None
 
+    large_index = max([atom_1_idx, atom_2_idx])
+    small_index = min([atom_1_idx, atom_2_idx])
+
+    shortest_path = atom_list[small_index:large_index + 1]
+    if len(shortest_path) - 1 > max_distance:
+        shortest_path = atom_list[large_index:] + atom_list[:small_index + 1]
+
+    return shortest_path
